/**
 * Polinomio.java
 *
 * Modificado el 16 de diciembre de 2004, 12:35
 * Modificado el 14 de enero de 2005, 22:15
 * Modificado en enero/febrero de 2006
 * Modificado en abril de 2007
 */


package auxiliar;
import ORG.netlib.math.complex.Complex;
import java.util.*; // Este paquete se necesita para usar el objeto Vector

/**
 *
 * @author  Alumnos de "Mtodos Numricos" desde 2002-2003,
 *          Antonio Pallars, Vctor Jimnez
 *
 * MTODOS NUMRICOS
 * LICENCIATURA DE MATEMTICAS
 * UNIVERSIDAD DE MURCIA
 */

/**
 * Cada objeto polinomio p(x)=a0 + a1 x + a2 x^2 + ... + ak x^k
 * tiene como atributo una sucesin de nmeros complejos
 * con sus coeficientes {a0,a1,...,ak}
 */
public class Polinomio {
    
    private Complex[] coef;
    static double precision = 5E-16;
    
    /** 
     * Constructor de un objeto Polinomio. <p>
     * Por defecto es el polinomio nulo.
     */
    public Polinomio() {
        coef = new Complex[1];
        coef[0] = new Complex();
    }
    
    /** 
     * Sobrecarga del Constructor de un objeto Polinomio. <p>
     * Polinomio(coeficientes) crea el polinomio cuyos coeficientes son 
     * los reales de la lista "coeficientes".
     */
    public Polinomio(double[] coeficientes) {
        int n = coeficientes.length;
        coef = new Complex[n];
        
        /** 
         * Comenzamos eliminando los ceros superfluos de la sucesin 
         * de coeficientes
         */
        for(int j = n - 1; j > 0; j--) {
            if (Math.abs(coeficientes[j]) < precision) {
                n = n - 1;
            } else {
                break;
            }
        }
        coef = new Complex[n];
        for(int i = 0; i < n; i++) {
            coef[i] = new Complex(coeficientes[i]);
        }
    }
    
    /** 
     * Sobrecarga del Constructor de un objeto Polinomio. <p>
     * Polinomio(coeficientes) crea el polinomio cuyos coeficientes 
     * son los complejos de la lista "coeficientes".
     */
    public Polinomio(Complex[] coeficientes) {
        int n = coeficientes.length;
                
        for(int j = n - 1; j > 0; j--) {
            if (coeficientes[j].abs() < precision) { 
                n = n - 1;
            } else {
                break;
            }
        }
        coef = new Complex[n];
        for(int i = 0; i < n; i++) {
            coef[i] = new Complex(coeficientes[i]);
        }
    }
    
    /** 
     * Sobrecarga del Constructor de un objeto Polinomio. <p>
     * Polinomio(q) crea un polinomio con los mismos coeficientes que q.
     */
    public Polinomio(Polinomio q) {
        int n = q.coef.length;
        coef = new Complex[n];
        
        for(int i = 0; i < n; i++){
            coef[i] = new Complex(q.coef[i]);
        }
    }
    
    /** 
     * Sobrecarga del Constructor de un objeto Polinomio. <p>
     * Polinomio(z) crea un polinomio constante igual al complejo z.
     */
    public Polinomio(Complex z) {
        coef = new Complex[1];
        
        coef[0] = z;
    }
    
    /** 
     * Sobrecarga del Constructor de un objeto Polinomio. <p>
     * Polinomio(x) crea un polinomio constante igual al real x.
     */
    public Polinomio(double x) {
        coef = new Complex[1];
        
        coef[0] = new Complex(x, 0);
    }
        
    /* Mtodo para obtener el grado del polinomio */
    public int grado() {
        return coef.length - 1;
    }
    
    public Complex[] coeficientes() {
        return coef;
    }
    
    /**
     * Mtodo para escribir un polinomio. <p>
     * Construye una cadena de caracteres (String) en la
     * que aparece escrito el polinomio.
     */
    public String escribe() {
        Complex[] y = coef;
        int m = y.length;
        String cadena = "( ";
        String renglon = "";
        String aux;
        
        for(int k = 0; k < m; k++) {
            if( (m > 1) && (y[k].abs() < precision) ) {
                continue; // evita escribir sumandos nulos
            }
            aux = "";
            if(k == 0) {
                aux = ""; // trmino independiente
            } else if (k == 1) {
                aux = " x"; // primera potencia
            } else {
                aux = " x^" + k; // resto de potencias
            }
            if(k == m - 1) {
                aux = aux + " )";
            } else {
                aux = aux + " + ";
            }
            aux = y[k].toStringMejorado() + aux;
            if((renglon + aux).length() > 90) {
                cadena = cadena + renglon + "\n  ";
                renglon = aux;
            } else {
                renglon = renglon + aux;
            }
        }
        return cadena + renglon;
    }
    
    /** 
     * Sobrecarga del mtodo toString para que pueda escribirse directamente
     * el polinomio
     */
    public String toString() {
        return escribe();
    }
    
    /**
     * Mtodo para escribir el desarrollo de Taylor de un
     * polinomio complejo centrado en un nmero complejo
     */
    public String escribeTaylor(Complex z) {
        Complex[] y = taylor(z);
        int m = y.length;
        String cadena = "( ";
        String renglon = "";
        String aux;
        
        if (z.abs() < precision) {
            return escribe();
        }
        for(int k = 0; k < m; k++) {
            if( (m > 1) && (y[k].abs() < precision) ) {
                continue;
            }
            aux = "";
            if(k == 0){
                aux = "";
            } else if (k == 1) {
                if (Math.abs(z.im()) < precision) {
                    if (z.re() < 0) {
                        aux = " (x + " + (- z.re()) + ")";
                    } else {
                        aux = " (x - " + z.re() + ")";
                    }
                } else {
                    aux = " (x - " + z + ")";
                }
            } else {
                if (Math.abs(z.im()) < precision) {
                    if (z.re() < 0) {
                        aux = " (x + " + (- z.re()) + ")^" + k;
                    } else {
                        aux = " (x - " + z.re() + ")^" + k;
                    }
                } else {
                    aux = " (x - " + z + ")^" + k;
                }
            }
            if(k == m - 1) {
                aux = aux + " )";
            } else {
                aux = aux + " + ";
            }
            aux = y[k].toStringMejorado() + aux;
            if((renglon + aux).length() > 90) {
                cadena = cadena + renglon + "\n  ";
                renglon = aux;
            } else {
                renglon = renglon + aux;
            }
        }
        return cadena + renglon;
    }
    
    /**
     * Sobrecarga del mtodo anterior para un polinomio real
     * centrado en un nmero real
     */
    public String escribeTaylor(double x) {
        Complex z = new Complex(x);
        
        return escribeTaylor(z);
    }
  
    /* Mtodo para obtener la suma de p(x) + q(x) */
    public Polinomio suma(Polinomio q) {
        int n = coef.length;
        int m = q.coef.length;
        int j = 0;
        Complex[] suma;
        
        if (n < m) {
            suma = new Complex[m];
            while(j < n) {
                suma[j] = coef[j].add(q.coef[j]);
                j++;
            }
            for(int i = j; i < m; i++) {
                suma[i] = q.coef[i];
            }
        } else {
            suma = new Complex[n];
            while(j < m) {
                suma[j] = coef[j].add(q.coef[j]);
                j++;
            }
            for(int i = j; i < n; i++) {
                suma[i] = coef[i];
            }
        }
        return new Polinomio(suma);
    }
    
    /** 
     * Mtodo para multiplicar dos polinomios <p>
     *
     * DATOS DE ENTRADA <p>
     * Polinomio q <p>
     *
     * UTILIZACIN <p>
     * p.producto(q) proporciona el Polinomio pq
     */
    public Polinomio producto(Polinomio q) {
        Complex[] pq;
        int n = coef.length;
        int m = q.coef.length;
        int k = m + n - 1;
        pq = new Complex[k];
        
        for(int i = 0; i < k; i++){
            pq[i] = new Complex(0);
            for(int j = 0; j <= i; j++){
                if (j >= n){ 
                    continue;
                }
                if ((i-j) >= m){ 
                    continue;
                }
                pq[i] = pq[i].add( coef[j].mul(q.coef[i-j]));
            }
        }
        return new Polinomio(pq);
    }
    
    /** 
     * Sobrecarga del mtodo producto para multiplicar 
     * un polinomio por un escalar <p>
     *
     * DATOS DE ENTRADA <p>
     * Complex z <p>
     *
     * UTILIZACIN <p>
     * p.producto(z) proporciona el Polinomio pq
     */
    public Polinomio producto(Complex z) {
        Complex[] pq;
        int n = coef.length;
        pq = new Complex[n];
        
        for(int i = 0; i < n; i++) {
            pq[i] = z.mul(coef[i]);
        }
        return new Polinomio(pq);
    }
    
    /* Mtodo para calcular el polinomio derivada */
    public Polinomio derivada() {
        Complex[] dp;
        int n = coef.length;
        
        if (n == 1) {
            return new Polinomio();
        } else {
            dp = new Complex[n - 1];
            for(int i = 0; i < n - 1; i++) {
                dp[i] = coef[i + 1].mul(new Complex(i + 1));
            }
            return new Polinomio(dp);
        }
    }
    
    
    /**
     * Sobrecarga del mtodo producto para multiplicar 
     * un polinomio por un escalar real
     */
    public Polinomio producto(double x) {
        Complex z = new Complex(x, 0);
        
        return producto(z);
    }
   
    /* Mtodo para restar dos polinomios */
    public Polinomio resta(Polinomio q) {
        Polinomio p = new Polinomio(coef);
        
        return p.suma(q.producto(-1));
    }
    
    /** 
     * Mtodo para dividir dos polinomios <p>
     *
     * DATOS DE ENTRADA <p>
     * Polinomio q <p>
     *
     * UTILIZACION <p>
     * p.div(q) proporciona el resto y el cociente de p/q 
     */
    public Polinomio[] div(Polinomio divisor) throws ErrorPolinomios {                  
        Polinomio[] div = new Polinomio[2]; // div[0] contendr el resto
                                          // div[1] contendr el cociente
        int n = coef.length; // grado del dividendo + 1
        int m = divisor.coef.length; // grado del divisor + 1
        int k = n - m + 1; 
        Complex[] aux = new Complex[n];
        Complex[] resto = new Complex[n]; // lista para guardar los restos
        Complex[] cociente = new Complex[k]; // lista para guardar los cocientes
                
        if (divisor.coef[m - 1].abs() == 0) {
            throw new ErrorPolinomios(); // el divisor es nulo
        }
        if (m > n) {
            div[1] = new Polinomio(); // el cociente es cero
            div[0] = new Polinomio(coef);  // el resto es el dividendo
            return div;
        }
        if (m == 1) {
            div[0] = new Polinomio(); // el resto es 0            
            for(int j = 0; j < n; j++) {
                aux[j] = coef[j].div(divisor.coef[0]);
            }
            div[1] = new Polinomio(aux);
            return div;
        }       
        for (int j = 0; j <= n - 1; j++) {
            resto[j] = coef[j];
        }
        for (int j = n - m; j >= 0; j--) {
            cociente[j] = resto[m + j - 1].div(divisor.coef[m - 1]);
            resto[m + j - 1] = new Complex(0);
            for(int h = m - 2; h >= 0; h--) {
                resto[j + h] = 
                        resto[j + h].sub(divisor.coef[h].mul(cociente[j]));
            }
        }
        div[0] = new Polinomio(resto);
        div[1] = new Polinomio(cociente);
        return div;        
    }            

    /** 
     * Mtodo de evaluacin de polinomios 
     * con coeficientes y variable compleja
     */
    public Complex eval(Complex z) {
        int n = grado();
        Complex z2 = new Complex(0);
        
        while (n != 0) {
            z2 = z2.add(coef[n]); // z2 = z2+coef[n]
            z2 = z2.mul(z); // z2 = z2*z
            n = n - 1;
        }
        z2 = z2.add(coef[0]); //z2 = z2+z
        return z2;
    }
    /** 
     * Sobrecarga del mtodo eval para evaluar polinomios reales
     * en variables reales
     */
    public double eval(double x) {
        Complex z = new Complex(x);
        
        return eval(z).re();
    }
    
    /** 
     * Mtodo de deflacin de polinomios
     * con coeficientes y variable compleja
     */
    public Polinomio deflacion(Complex z) {
        int n = grado();
        Complex defl[] = new Complex[n];
        int k = n - 1;
        
        if (n == 0){
            return new Polinomio(coef); // el polinomio a deflacionar es constante
        }        
        defl[n - 1]=coef[n];
        while (k >= 1) {
            defl[k - 1] = z.mul(defl[k]);
            defl[k - 1] = coef[k].add(defl[k - 1]);
            k = k - 1;
        }
        return new Polinomio(defl);
    }
    
    /** 
     * Sobrecarga del mtodo deflacion para deflacionar polinomios reales
     * en variables reales
     */
    public Polinomio deflacion(double x) {
        Complex z = new Complex(x);
        
        return new Polinomio(deflacion(z));
    }
    
    /**
     * Mtodo que calcula los coeficientes de Taylor
     * de un polinomio complejo centrado en un nmero complejo.
     */
    public Complex[] taylor(Complex z) {
        int n = grado();
        Complex[] coefTaylor = new Complex[n + 1];
        Complex[] aux = new Complex[n + 1];
        
        for (int i = 0; i <= n; i++) {
            aux[i] = coef[i];
        }
        coefTaylor[n] = aux[n];
        for (int i = 0; i <= n - 1; i++) {
            for (int k = n - 1; k >= i + 1; k--) {
                coefTaylor[k] = z.mul(coefTaylor[k + 1]).add(aux[k]);
                aux[k] = coefTaylor[k];
            }
            coefTaylor[i] = z.mul(coefTaylor[i + 1]).add(aux[i]);
        }
        return coefTaylor;
    }
    
    /**
     * Sobrecarga del mtodo anterior para un polinomio real
     * centrado en un nmero real.
     */
    public double[] taylor(double x) {
        Complex z = new Complex(x);
        int n = grado();
        double[] coefTaylor = new double[n + 1];
        
        for (int i = 0; i <= n; i++) {
            coefTaylor[i] = taylor(z)[i].re();
        }
        return coefTaylor;
    }
    
    /** 
     * Mtodo de evaluacin de un polinomio y su derivada
     * con coeficientes y variable compleja
     */
    public Complex[] horner2Complejo(Complex z) {
        int n = grado();
        Complex[] valores = new Complex[2];
        valores[0] = new Complex(0);
        valores[1] = new Complex(0);
        
        while (n != 1) {
            valores[0] = z.mul(valores[0].add(coef[n]));
            valores[1] = z.mul(valores[1].add(coef[n].mul(new Complex(n))));
            n = n - 1;
        }
        valores[0] = z.mul(valores[0].add(coef[1]));
        valores[1] = valores[1].add(coef[1]);
        valores[0] = valores[0].add(coef[0]);
        return valores;
    }
        
    /**
     * Mtodo para obtener una cota superior de las races
     * de un polinomio.
     */    
    public double cotaRaices() {
        int n = coef.length;
        double max;
                
        if (n == 0) {
            return 0;
        }  else {
            max = 0;
            for (int i = 0; i < n - 1; i++) {
                if (coef[i].abs() > max) {
                    max = coef[i].abs();
                }
            }
            return 1 + max / coef[n - 1].abs();
        } 
    }    
    
    /** 
     * MTODO DE NEWTON PARA POLINOMIOS <p>
     *
     * DATOS DE ENTRADA: <p>
     * Complex xi: aproximacin inicial xi a la raz del polinomio <p>
     * double tol: precisin real tol para la condicin de parada <p>
     * int nmax: nmero mximo de iteraciones <p>
     * boolean imp: variable lgica para s/no imprimir los pasos intermedios <p>
     *
     * DATOS DE SALIDA: <p>
     * Complex: aproximacin a la raz del polinomio con precisin menor que tol
     *      y en menos de nmax pasos <p>
     * Polinomios.ErrorPolinomios: excepcin a capturar para el caso en el 
     *      que no haya convergencia o la derivada del polinomio se 
     *      anule en algn paso 
     */    
    public Complex newton(Complex xi, double tol, int nmax, boolean imp)
            throws ErrorPolinomios {
        Complex u[] = new Complex[2];
        Complex xf = new Complex(0);
        String mensaje = "";
        
        u = horner2Complejo(xi); // u[0] = p(xi); u[1] = p'(xi);
        if (u[0].abs() < tol) {
            if (imp == true) {
                mensaje = mensaje + "\nEl resultado es " 
                        + xi.toStringMejorado();
                System.out.println("\nNewton:" + mensaje);                
            }
            return xi;
        }
        if (u[1].abs() < precision) {
            mensaje = mensaje + "\nError: derivada nula en " 
                    + xi.toStringMejorado();
            System.out.println("Newton:" + mensaje);                
            throw new ErrorPolinomios();
        }        
        for (int k = 1; k <= nmax; k++) {
            xf = xi.sub(u[0].div(u[1]));
            u = horner2Complejo(xf); // u[0] = p(xf); u[1] = p'(xf);
            if (imp == true) {
                mensaje = mensaje + "\nPaso " + k
                        + ": xi = " + xi.toStringMejorado() +  "; p(xi) = " 
                        + u[0].toStringMejorado() + "; \n  p'(xi) = " 
                        + u[1].toStringMejorado()
                        + "; xf = " + xf.toStringMejorado();
            }
            if (u[0].abs() < tol) {
                if (imp == true) {                 
                    mensaje = mensaje + "\nRaz en " + xf.toStringMejorado() 
                            + " en " + k + " iteraciones";
                    System.out.println("\nNewton:" + mensaje);
                }
                return xf;
            }
            if (xf.sub(xi).abs() < tol) {
                if (imp == true) {                 
                    mensaje = mensaje + "\nRaz en " + xf.toStringMejorado() 
                            + " en " + k + " iteraciones";                
                    System.out.println("\nNewton:" + mensaje);
                }
                return xf;
            }
            if (u[1].abs() < tol){
                mensaje = mensaje + "\nLa derivada se anula en el paso " + k;
                System.out.println("Newton:" + mensaje);                                
                throw new ErrorPolinomios();
            }
            xi = xf;
        }
        mensaje = mensaje + "\nNo hay convergencia en " + nmax + " iteraciones";
        System.out.println("Newton:" + mensaje);                                
        throw new ErrorPolinomios();
    }
    
    /** 
     * Sobrecarga del mtodo de Newton para polinomios reales equivalente a
     *          Newton(new Complex(xi), tol, nmax, imprime).re()
     */
    public double newton(double xi, double tol, int nmax, boolean imprime)
            throws ErrorPolinomios {
        try {
            return newton(new Complex(xi), tol, nmax, imprime).re();
        } catch(ErrorPolinomios error) {
            throw new ErrorPolinomios();
        }
    }
    
    /** 
     * Sobrecarga del mtodo de Newton con complejos equivalente a
     *          Newton(xi, tol, nmax, false)
     */
    public Complex newton(Complex xi, double tol, int nmax)
            throws ErrorPolinomios {
        try {
            return newton(xi, tol, nmax, false);
        } catch(ErrorPolinomios error) {
            throw new ErrorPolinomios();
        }
    }
    
    /** 
     * Sobrecarga del mtodo de Newton para polinomios reales equivalente a
     *          Newton(new Complex(xi), tol, nmax, false).re()
     */
    public double newton(double xi, double tol, int nmax)
            throws ErrorPolinomios {
        try {
            return newton(new Complex(xi), tol, nmax, false).re();
        } catch(ErrorPolinomios error) {
            throw new ErrorPolinomios();
        }
    }
    
    /**
     * MTODO DE LAGUERRE <p>
     *
     * DATOS DE ENTRADA: <p>
     * Complex z[]: aproximacin inicial z a la raz del polinomio <p>
     * double tol: precisin real para la condicin de parada <p>
     * int nmax: nmero mximo de realizaciones <p>
     *
     * DATOS DE SALIDA: <p>
     * raz aproximada o mensaje de error
     */
    public Complex laguerre(Complex z, double tol, int nmax, boolean imprime)
            throws ErrorPolinomios {
        Complex zini = new Complex();
        Complex zfinal = new Complex();
        Complex aux1 = new Complex();
        Complex aux2 = new Complex();
        Complex c1 = new Complex();
        Complex c2 = new Complex();
        zini = z;
        zfinal = z.add(new Complex(1));
        Complex b = new Complex();
        Complex c = new Complex();
        Complex d = new Complex();
        String mensaje = "";
        int n = coef.length -1;
        
        for(int j = 1; j <= nmax; j++) {
            
            /* Evaluacion de p(z), p'(z), p''(z) mediante el mtodo de Horner */
            b = coef[n];
            c = b;
            d = b;
            for(int k = n - 1; k >= 2; k--){
                b = b.mul(zini).add(coef[k]); // b = bz + (coeficiente k-simo)
                c = c.mul(zini).add(b); // c = cz+b para calcular la derivada
                d = d.mul(zini).add(c); // d = dz+c para calcular p''(z)=2d
            }
            b = b.mul(zini).add(coef[1]); // b = bz + (primer coeficiente)
            c = c.mul(zini).add(b); // c = cz+b obteniendo as p'(z)=c
            b = b.mul(zini).add(coef[0]); // b = bz + (trmino independiente)
                                          // obteniendo as p(z)=b
            if(imprime == true) {
                mensaje = mensaje + "\nPaso " + j
                        + ": z = " + zini.toStringMejorado() + "; p(z) = " 
                        + b.toStringMejorado();
            }
            // final de la evaluacin 
            
            /* 1 condicin de parada */
            if (b.abs() < tol) {
                if (imprime == true) {
                    mensaje = mensaje + "\nRaz en z = " 
                            + zini.toStringMejorado() + "; p(z) = " 
                            + b.toStringMejorado() + " en " + j 
                            + " iteraciones";
                    System.out.println("\nLaguerre:" + mensaje);
                }
                return zini;
            }
            
            /* Construccion de nueva aproximacin */
            aux1 = new Complex(0);
            aux1 = aux1.pow(c.mul(new Complex(n-1)),2); // [[(n-1)^2]c^2]
            aux1 = aux1.sub(d.mul(b).mul(new Complex(n-1)).mul(new Complex(n)));
                    // [[(n-1)^2]c^2]-n(n-1)bd
            aux1 = aux1.sqrt();
            aux2 = new Complex(1);
            aux2 = aux2.div(b.mul(new Complex(n))); // (1)/nb
            c1= aux2.mul(c.add(aux1));
                    // [(1)/nb]*[c+[[[(n-1)^2]c^2]-n(n-1)bd]^1/2]
            c2= aux2.mul(c.sub(aux1));
                    // [(1)/nb]*[c-[[[(n-1)^2]c^2]-n(n-1)bd]^1/2]
            
            /* Elegimos c1 o c2 segn el que proporcione mayor valor absoluto */
            if (c1.abs() < c2.abs()) {
                c1 = c2;
            }
            
            /* 1 condicin de parada sin convergencia */
            if (c1.abs() < 1.0E-16) {
                mensaje = mensaje + "\nError: primera y segunda "
                        + "derivada casi-nulas";
                System.out.println("Laguerre:" + mensaje);
                throw new ErrorPolinomios();
            }
            
            /* La nueva iterada se obtiene restando 1/c1 a la condicin inicial */
            aux1 = new Complex(1).div(c1);
            zfinal = zini.sub(aux1);
            
            /* 2 condicin de parada */
            if (aux1.abs() < tol) {  
                b = coef[n];            
                for(int k = n - 1; k >= 0; k--){
                    b = b.mul(zfinal).add(coef[k]); 
                }
                if (imprime == true) {
                    mensaje = mensaje + "\nPaso " + (j + 1)
                            + ": z = " + zfinal.toStringMejorado() 
                            + "; p(z) = " + b.toStringMejorado();
                    mensaje = mensaje + "\nRaz en z = " 
                            + zfinal.toStringMejorado() + "; p(z) = " + 
                            b.toStringMejorado() + " en " + (j + 1) 
                            + " iteraciones";
                    System.out.println("\nLaguerre:" + mensaje);
                }
                return zfinal;
            }            
            zini = zfinal;
        }
        // Concluye el paso y se inicia el siguiente
        
        /* 2 condicin de parada sin convergencia */       
        mensaje = mensaje + "\nError: no hay convergencia en " 
                + nmax + " etapas";
        System.out.println("Laguerre:" + mensaje);
        throw new ErrorPolinomios();
    }
    
    /** 
     * Sobrecarga del mtodo de Laguerre para polinomios reales equivalente a 
     *      laguerre(new Complex(xi), tol, nmax, imprime).re()
     */
    public double laguerre(double xi, double tol, int nmax, boolean imprime)
            throws ErrorPolinomios {
        try {
            return laguerre(new Complex(xi), tol, nmax, imprime).re();
        } catch(ErrorPolinomios error) {
            throw new ErrorPolinomios();
        }
    }
    
    /** 
     * Sobrecarga del mtodo de Laguerre con complejos equivalente a
     *      laguerre(xi, tol, nmax, false)
     */
    public Complex laguerre(Complex xi, double tol, int nmax)
            throws ErrorPolinomios {
        try {
            return laguerre(xi, tol, nmax, false);
        } catch(ErrorPolinomios error) {
            throw new ErrorPolinomios();
        }
    }
    
    /** 
     * Sobrecarga del mtodo de Laguerre para polinomios reales equivalente a
     *      laguerre(new Complex(xi), tol, nmax, false).re()
     */
    public double laguerre(double xi, double tol, int nmax)
            throws ErrorPolinomios {
        try {
            return laguerre(new Complex(xi), tol, nmax, false).re();
        } catch(ErrorPolinomios error) {
            throw new ErrorPolinomios();
        }
    }

    /** 
     * Mtodo para medir el signo de un nmero real.
     * Utiliza  el mtodo Double.isNaN() para saber si el nmero es cero.
     */
    public int signo(double x) {
        double test = x / Math.abs(x);
        
        if (Double.isNaN(test)) {
            return 0;
        } else {
            return (int) test;
        }
    }
        
    /** 
     * Mtodo para contar cambios de signo en listas de reales.
     * Devuelve el nmero de cambios de signo.
     * La determinacin de signos en nmeros prximos a cero es muy inestable.
     * Pueden aparecer problemas cuando aparecen ceros en los trminos.
     */
     public int cuentaCambiosDeSigno(double[] lista) {
        int ncambios = 0;
        int ini = 0; 
        int sig = 0;
        int m = lista.length;
        double pini = lista[ini];
        double psig = 0.;
        
        while (ini < (m - 1)) {
            if (signo(pini) == 0) { // pini == 0
                ini = ini + 1;
                pini = lista[ini];
            } else {
                break;
            }
        }
        if (ini == (m - 1)) {
            return ncambios;
        }
        while (ini < m - 1) {
            sig = ini + 1;
            while (sig < m) {
                psig = lista[sig];
                if(signo(psig) == 0) { // psig == 0
                    sig = sig + 1;
                } else { 
                    break;
                }
            }
            if (signo(pini) != signo(psig)) {
                ncambios = ncambios + 1;
            }
            ini = sig;
            pini = psig;
        }
        return ncambios;
    }
     
    /* Mtodo para construir la sucesion de polinomios de Sturm */    
    public Polinomio[] sucesionSturm(boolean imprime) {
        int n = coef.length;
        Polinomio[] sturm = new Polinomio[n];
        sturm[0] = new Polinomio(coef);
        double[] b;
        
        if (n == 1) {
            return sturm;
        }
        b = new double[n - 1];
        for (int i = 0; i < n - 1; i++) {
            b[i] = coef[i + 1].mul(new Complex(i + 1)).re();
        }
        sturm[1] = new Polinomio(b);
        if (n == 2) {
            return sturm;
        }
        try {
            for (int k = 2; k < n; k++) {
                Polinomio auxiliar =  sturm[k - 2].div(sturm[k - 1])[0];
                if (auxiliar.grado() == 0) {
                    if (auxiliar.coef[0].abs() < precision) {
                        for (int j = 0; j <= k - 1; j++) {
                            sturm[j] = sturm[j].div(sturm[k - 1])[1];
                            if (imprime == true) {
                                System.out.println("Paso " + j 
                                        + "; polSturm = "+ sturm[j].escribe());
                            }
                        }
                        return sturm;
                    }
                    sturm[k] = auxiliar.producto(new Complex(- 1));
                    if (imprime == true) {
                        for (int j = 0; j <= k ; j++) {
                            System.out.println("Paso " + j 
                                    + "; polSturm = " + sturm[j].escribe());
                        }
                    }
                    return sturm;
                }
                sturm[k] = auxiliar.producto(new Complex(- 1));
            }
        } catch(ErrorPolinomios err) {
        }
        if (imprime == true) {
            for (int j = 0; j <= n - 1; j++) {
                System.out.println("Paso " + j + "; polSturm = " 
                        + sturm[j].escribe());
            }
        }
        return sturm;        
    }
    
    /** 
     * Sobrecarga del mtodo sucesionSturm partiendo de un polinomio
     * con coeficientes los del vector a     
     */
    public Polinomio[] sucesionSturm(double a[], boolean imprime) {
        Polinomio p = new Polinomio(a);
        
        return p.sucesionSturm(imprime);
    }
   
    /* Sobrecarga del mtodo sucesionSturm. Equivale a sucesionSturm(false). */ 
    public Polinomio[] sucesionSturm() {
        return sucesionSturm(false);
    }
    
    /* Sobrecarga del mtodo sucesionSturm. Equivale a sucesionSturm(a, false). */
    public Polinomio[] sucesionSturm(double a[]) {
        return sucesionSturm(a, false);
    }
    
    /** 
     * MTODO DE STURM <p>
     *
     * PROGRAMA RECURSIVO QUE USA COMO DATOS DE ENTRADA: <p>
     * double prec: precisin para aproximar  races reales <p>
     * Vector lista : un vector donde almacenar los intervalos que 
     *      contienen una raz y tienen longitud menor que prec <p>
     * Polinomio[] sturm: los polinomios de la sucesin de Sturm <p>
     * double a y double b: los extremos del intervalo inicial <p>
     *
     * COMO DATOS DE SALIDA: <p>
     * void: no devuelve ningun dato; modifica el vector lista que le pasamos 
     *      como parametro, donde se han almacenado los intervalos de 
     *      longitud menor que prec y que contienen una raz obtenida 
     *      mediante biseccin
     */    
    public void bisecSturm(double prec, Vector lista, Polinomio[] sturm, 
            double a, double b) throws ErrorPolinomios {        
        int m = sturm.length;
        double[] sb = new double[m];
        double[] sa = new double[m];
        double[] interv = {a,b};
        
        for(int i = 0; i < m; i++) {
            sb[i] = sturm[i].eval(b);
            sa[i] = sturm[i].eval(a);
        }
        int n = cuentaCambiosDeSigno(sa) - cuentaCambiosDeSigno(sb);
        if (n <= 0) {
            return;
        }
        
        /**
         * La ultima condicin es por controlar el caso de raices multiples
         * y la inestabilidad del calculo de los cambios de signo
         */        
        if ( ((n == 1) && (Math.abs(b - a) < prec)) || 
                (Math.abs(b - a) < 1000 * precision) ) {            
            lista.addElement(interv);
            return;            
        } else {
            double c = a + 0.5 * (b - a);            
            bisecSturm(prec, lista, sturm, a, c);
            bisecSturm(prec, lista, sturm, c, b);
            return;
        }      
    }
    
    /** 
     * MTODO BISECCIN DE STURM CON PRECISIN <p>
     *
     * PROGRAMA QUE USA COMO DATOS DE ENTRADA: <p>
     * double prec: precisin para aproximar  races reales <p>
     * double a y double b: los extremos del intervalo inicial <p>
     *
     * COMO DATOS DE SALIDA: <p>
     * Nos devuelve una lista con los extremos de los intervalos donde
     * se localizan las raices reales, teniendo estos intervalos 
     * longitud menor que prec
     */
    public double[][] bisecSturm(double prec) throws ErrorPolinomios {
        
        /**
         * La clase Vector se necesita para utilizar listas
         * a las que se les pueden aadir o quitar elementos
         */
        Vector lista = new Vector();         
        Polinomio[] sturm = sucesionSturm();
        int n = coef.length;
        double a;
        double b;
        double c;
        int longitud;
        double[][] lista2;
        
        if (n == 1) {
            System.out.println("Error: polinomio casi-constante");
            throw new ErrorPolinomios();
        }        
        b = 0;
        for(int k = 0; k < n; k++) {
            c = coef[k].abs();
            if(c > b) {
                b = c;
            }
        }
        b = 1 + b / coef[n-1].abs();
        a = - b;
        bisecSturm(prec, lista, sturm, a, b);
        
        /**
         * El Vector lista contiene los intervalos con las races;
         * vamos a convertirlo en una matriz de numeros reales
         */        
        longitud = lista.size();
        if (longitud == 0) {
            System.out.println("Error: no hay ceros reales");
            throw new ErrorPolinomios();
        }
        lista2 = new double[longitud][2];        
        for(int i = 0; i < longitud; i++) {
            lista2[i] = (double[]) lista.get(i);
        }        
        return lista2;        
    }    
    
    /* Exception para ser lanzada por los distintos mtodos */
    public static class ErrorPolinomios extends Exception {
        ErrorPolinomios() {
            super("Error: polinomios");
        }
    }
    
    
}
